@@ -1,10 +1,13 @@
-from __future__ import unicode_literals
+from __future__ import unicode_literals, print_function
 import os
 import sys
 import six
+import time
 import types
-import logging
+import base64
 import shutil
+import inspect
+import logging
 import functools
 import posixpath
 from subprocess import Popen
@@ -13,6 +16,23 @@ from subprocess import STDOUT
 from copy import deepcopy
 from contextlib import contextmanager
 
+__all__ = [
+    'console_script', 'sh', 'env', 'stdin', 'test',
+    'ls', 'cat', 'grep', 'find', 'cut', 'tr', 'head', 'tail', 'sed', 'awk',
+    'nc', 'ping', 'nmap', 'hostname', 'host', 'scp', 'rsync', 'wget', 'curl',
+    'cd', 'which', 'mktemp', 'echo', 'wc',
+    'tar', 'gzip', 'gunzip', 'zip', 'unzip',
+    'vlc', 'ffmpeg', 'convert',
+    'virtualenv', 'pip',
+    'ssh', 'sudo',
+    'path', 'pwd',  # path is posixpath, pwd return os.getcwd()
+    'escape', 'e',  # e is escape()
+]
+
+__not_piped__ = ['chmod', 'cp', 'mkdir', 'mv', 'rm', 'rmdir', 'touch']
+
+__all__ += __not_piped__
+
 log = logging.getLogger('chut')
 
 aliases = dict(
@@ -22,30 +42,6 @@ aliases = dict(
   )
 
 
-def console_script(func):
-    @functools.wraps(func)
-    def wrapper(arguments=None):
-        doc = getattr(func, '__doc__', None)
-        if doc is None:
-            doc = 'Usage: %prog'
-        name = func.__name__.replace('_', '-')
-        doc = doc.replace('%prog', name).strip()
-        doc = doc.replace('\n    ', '\n')
-        # take care if a script is chutified
-        if 'docopt' not in sys.modules:
-            import docopt
-        else:
-            docopt = sys.modules['docopt'] # NOQA
-        if isinstance(arguments, list):
-            arguments = docopt.docopt(doc, args=arguments)
-            return func(arguments)
-        else:
-            arguments = docopt.docopt(doc, help=True)
-            sys.exit(func(docopt.docopt(doc)))
-    wrapper.console_script = True
-    return wrapper
-
-
 def check_sudo():
     sudo = aliases.get('sudo')
     if not os.path.isfile(sudo):
@@ -56,15 +52,29 @@ def check_sudo():
     whoami = Popen(args, env=env, **kwargs)
     whoami.wait()
     whoami = whoami.stdout.read().strip()
-    if whoami != 'root':
-        raise OSError('Not able to run sudo')
+    if whoami != six.b('root'):
+        raise OSError('Not able to run sudo.')
+
+
+def escape(value):
+    chars = "|!`'[]() "
+    esc = '\\'
+    if isinstance(value, six.binary_type):
+        chars = chars.encode('ascii')
+        esc = esc.encode('ascii')
+    for c in chars:
+        value = value.replace(c, esc + c)
+    return value
 
 
 class Environ(dict):
     """Manage os.environ"""
 
     def __getattr__(self, attr):
-        return self.get(attr.upper(), None)
+        value = self.get(attr.upper(), None)
+        if attr.lower() in ('path',):
+            return value.split(os.pathsep)
+        return value
 
     def __setattr__(self, attr, value):
         if isinstance(value, (list, tuple)):
@@ -84,26 +94,37 @@ class Pipe(object):
     _chut = None
     _pipe = True
     _cmd_args = []
+    _sys_stdout = sys.stdout
+    _sys_stderr = sys.stderr
 
     def __init__(self, *args, **kwargs):
+        self._done = False
+        self._stdout = None
         self._stderr = None
         self.args = list(args)
         self.previous = None
         self.processes = []
-        self.encoding = kwargs.get('encoding', 'utf8')
+        encoding = kwargs.get('encoding')
+        if not encoding:
+            encoding = getattr(sys.stdout, 'encoding', None) or 'utf8'
+        self.encoding = encoding
         self.kwargs = kwargs
         if 'sh' in kwargs:
             kwargs['shell'] = kwargs.pop('sh')
         if 'combine_stderr' in kwargs:
             kwargs.pop('combine_stderr')
             kwargs['stderr'] = STDOUT
-        if 'binary' in kwargs:
-            self._binary = kwargs.pop('binary')
         if 'pipe' in kwargs:
             if not kwargs.pop('pipe'):
-                self.__call__()
+                self._call_pipe()
         elif not self._pipe:
-            self.__call__()
+            self._call_pipe()
+
+    def _call_pipe(self):
+        self._done = True
+        ret = self.__call__()
+        if ret.failed:
+            print(ret.stderr, file=sys.stderr)
 
     @property
     def returncodes(self):
@@ -162,8 +183,10 @@ class Pipe(object):
         if self._cmd_args[:1] == ['ssh']:
             cmd = '%s %s' % (binary, ' '.join(self.args))
             cmd = cmd.strip()
-            if ('|' in cmd or '>' in cmd) and shell:
-                cmd = repr(str(cmd))
+            for c in '\'"*<>|':
+                if c in cmd:
+                    cmd = repr(str(cmd))
+                    break
             args.append(cmd)
         else:
             args.extend(binary.split())
@@ -227,12 +250,62 @@ class Pipe(object):
                 try:
                     p = Popen(args, **kwargs)
                 except OSError:
-                    self._raise(args, kwargs)
+                    self._raise()
 
                 self.processes.append(p)
                 stdin = p.stdout
         return stdin
 
+    @classmethod
+    def map(cls, args,
+            pool_size=None, stop_on_failure=False, **kwargs):
+        """Run a batch of the same command and manage a pool of processes for
+        you"""
+        kw = dict(
+            stdin=sys.stdin, stderr=PIPE,
+            stdout=PIPE
+            )
+        kw.update(kwargs)
+        if pool_size is None:
+            import multiprocessing
+            pool_size = multiprocessing.cpu_count()
+        results = [None] * len(args)
+        processes = []
+        index = 0
+        out_index = 0
+        while args or processes:
+            if args and len(processes) < pool_size:
+                a = args.pop(0)
+                if not isinstance(a, list):
+                    a = [a]
+                cmd = cls(*a)
+                a = cmd.command_line(cmd.kwargs.get('shell', False))
+                processes.append((index, cmd, Popen(a, **kw)))
+                index += 1
+            for i, cmd, p in processes:
+                result = p.poll()
+                if result is not None:
+                    output = Stdout(p.stdout.read())
+                    output.stderr = p.stderr.read()
+                    output.returncodes = [result]
+                    output.failed = bool(result)
+                    output.succeeded = not output.failed
+                    results[i] = output
+                    processes.remove((i, cmd, p))
+                    if out_index == i:
+                        out_index += 1
+                        yield results[i]
+                    if result > 0 and stop_on_failure:
+                        args = None
+                        for index, cmd, p in processes:
+                            if p.poll() is None:
+                                p.kill()
+                        cmd._raise(output=output)
+            time.sleep(.1)
+        if out_index < len(results):
+            yield results[out_index]
+            out_index += 1
+
     def __getitem__(self, item):
         if not isinstance(item, slice):
             raise KeyError('You can only use slices')
@@ -247,9 +320,11 @@ class Pipe(object):
 
     def __iter__(self):
         for line in self.stdout:
-            yield line.decode(self.encoding)
+            yield self._decode(line)
 
     def __call__(self, **kwargs):
+        if self._done and self._stdout is not None:
+            return self._stdout
         for cmd in self.commands:
             if kwargs.get('shell'):
                 cmd.kwargs['shell'] = True
@@ -261,11 +336,13 @@ class Pipe(object):
                 output = stdout.read().rstrip()
             else:
                 output = b''.join(list(stdout)).rstrip()
-            if not isinstance(output, six.text_type):
-                output = output.decode(self.encoding)
+            output = self._decode(output)
         else:
             output = ''
-        return self._get_stdout(output)
+        output = self._get_stdout(output)
+        if self._done:
+            self._stdout = output
+        return output
 
     __str__ = __call__
 
@@ -304,16 +381,33 @@ class Pipe(object):
                 cmds[i].previous = cmds[i - 1]
         return cmds
 
+    def _write_to(self, fd):
+        if not isinstance(self, PyPipe):
+            self.kwargs['stdout'] = fd
+            return self.__call__()
+        else:
+            for line in self.stdout:
+                fd.write(line)
+            return self._get_stdout('')
+
     def _write(self, filename, mode):
-        with open(filename, mode) as fd:
-            if not isinstance(self, PyPipe):
-                self.kwargs['stdout'] = fd
-                return self.__call__()
+        if isinstance(filename, int):
+            if filename == 2:
+                fd = self._sys_stderr
             else:
-                for line in self.stdout:
-                    fd.write(line)
-                return self._get_stdout('')
-        return None
+                fd = self._sys_stdout
+            output = self._write_to(fd)
+        else:
+            with open(filename, mode) as fd:
+                output = self._write_to(fd)
+        if output.failed:
+            self._raise(output=output)
+        return output
+
+    def _decode(self, output):
+        if six.PY3 and not isinstance(output, six.text_type):
+            output = output.decode(self.encoding)
+        return output
 
     def _get_stdout(self, stdout):
         if not six.PY3 and not isinstance(stdout, six.binary_type):
@@ -325,14 +419,14 @@ class Pipe(object):
         output.succeeded = not output.failed
         return output
 
-    def _raise(self, args, kwargs):
-        if isinstance(args, list):
-            args = ' '.join(args)
-        env_ = kwargs.pop('env')
-        log.debug('Error while running Popen(%r, **%r)',
-                  args, kwargs)
-        kwargs['env'] = env_
-        raise OSError(args)
+    def _raise(self, output=None):
+        if not log.handlers:
+            logging.basicConfig(stream=sys.stderr)
+        if output is not None:
+            if output.stderr:
+                log.error(output.stderr)
+            raise OSError(self.commands_line, output.stderr)
+        raise OSError(self.commands_line)
 
 
 class Stdin(Pipe):
@@ -401,12 +495,11 @@ class PyPipe(Pipe):
         return self.func(self.stdin)
 
     def __deepcopy__(self, *args):
-        return ch.wraps(self.func)
+        return sh.wraps(self.func)
 
 
 class Base(object):
-    not_piped = ['rm', 'mkdir', 'cp', 'touch', 'mv', 'scp', 'rsync']
-    not_piped = sorted([str(c) for c in not_piped])
+    not_piped = [str(c) for c in __not_piped__]
 
     def __init__(self, name, *cmd_args):
         self.__name__ = name
@@ -442,6 +535,8 @@ class Base(object):
 
 class Chut(Base):
 
+    path = posixpath
+
     def wraps(self, func):
         return type(func.__name__, (PyPipe,), {'func': staticmethod(func)})()
 
@@ -460,9 +555,18 @@ class Chut(Base):
         return pipe(*args, **kwargs)
 
     def cd(self, directory):
+        """Change the current directory"""
         if self.__name__ not in ('sh', 'sudo'):
             raise ImportError('You can only run cd in local commands')
+        directory = os.path.realpath(directory)
         os.chdir(directory)
+        env.pwd = directory
+
+    def pwd(self):
+        """return os.path.abspath(os.getcwd())"""
+        if self.__name__ not in ('sh', 'sudo'):
+            raise ImportError('You can only use pwd in local commands')
+        return os.path.abspath(os.getcwd())
 
     def stdin(self, value):
         return Stdin(value)
@@ -490,7 +594,13 @@ class SSH(Base):
     """A ssh server"""
 
     def join(self, *args):
-        return '%s%s' % (self, posixpath.join(*args))
+        p = posixpath.join(*args)
+        host = str(self)
+        quote = '"'
+        if isinstance(p, six.binary_type):
+            host = host.encode('ascii')
+            quote = quote.encode('ascii')
+        return host + quote + escape(p) + quote
 
     @property
     def host(self):
@@ -525,7 +635,10 @@ class ModuleWrapper(types.ModuleType):
 
     def __getattr__(self, attr):
         if attr == '__all__':
-            raise ImportError('You cant import things that does not exist')
+            if self.__name__ == 'chut':
+                return [str(c) for c in __all__]
+            else:
+                raise ImportError('You cant import things that does not exist')
         if hasattr(self.mod, attr):
             return getattr(self.mod, attr)
         else:
@@ -535,31 +648,140 @@ class ModuleWrapper(types.ModuleType):
 
 
 env = Environ(os.environ.copy())
-ch = Chut('sh')
+sh = Chut('sh')
 sudo = Chut('sudo', '-s')
 test = Command('test')
-
-try:
-    import fabric.operations
-except ImportError:
-    pass
-else:
-    def _run_command(command, *args, **kwargs):
-        if hasattr(command, '_chut'):
-            new_command = command.commands_line
-            if isinstance(new_command, property):
-                new_command = command().commands_line
-        else:
-            new_command = command
-        return fab_run_command(new_command, *args, **kwargs)
-    fab_run_command = fabric.operations._run_command
-    fabric.operations._run_command = _run_command
+e = escape
 
 
 def wraps_module(mod):
-    sys.modules['chut'] = ModuleWrapper(mod, ch, 'chut')
+    sys.modules['chut'] = ModuleWrapper(mod, sh, 'chut')
     sys.modules['chut.sudo'] = ModuleWrapper(mod, sudo, 'sudo')
 
+#####################
+# Script generation #
+#####################
+
+
+def console_script(*args, **docopts):
+    def _console_script(func, **docopts):
+        @functools.wraps(func)
+        def wrapper(arguments=None):
+            if 'help' not in docopts:
+                docopts['help'] = True
+            if 'doc' not in docopts:
+                doc = getattr(func, '__doc__', None)
+            else:
+                doc = docopts.pop('doc')
+            if doc is None:
+                doc = 'Usage: %prog'
+            name = func.__name__.replace('_', '-')
+            doc = doc.replace('%prog', name).strip()
+            doc = doc.replace('\n    ', '\n')
+            import docopt
+            if isinstance(arguments, list):
+                docopts['argv'] = arguments
+                arguments = docopt.docopt(doc, **docopts)
+                return func(arguments)
+            else:
+                arguments = docopt.docopt(doc, **docopts)
+                sys.exit(func(arguments))
+        wrapper.console_script = True
+        return wrapper
+    if not args:
+        def waiting_for_func(func):
+            return _console_script(func, **docopts)
+        return waiting_for_func
+    else:
+        return _console_script(args[0])
+
+
+def generate(filename, arguments=None):
+    if arguments is None:
+        arguments = {}
+    if not os.path.isfile(filename):
+        mod = __import__(filename, globals(), locals(), [''])
+        filename = mod.__file__
+        name = mod.__name__
+    else:
+        dirname = os.path.dirname(filename)
+        sys.path.insert(0, dirname)
+        name = inspect.getmodulename(filename)
+        mod = __import__(name)
+
+    filenames = []
+    for k, v in mod.__dict__.items():
+        if getattr(v, 'console_script', False) is True:
+            filenames.append(k)
+
+    if arguments.get('--list-entry-points'):
+        for name in filenames:
+            print(('%s = %s:%s' % (name.replace('_', '-'),
+                                   mod.__name__, name)))
+        return 0
+
+    dest = os.path.expanduser(arguments.get('--destination', 'dist/scripts'))
+    sh.mkdir('-p', dest)
+
+    def encode_module(mod):
+        if not hasattr(mod, '__file__'):
+            mod = __import__(mod)
+        data = inspect.getsource(mod)
+        data = base64.encodestring(six.b(data))
+        return 'mods.append((%r, %r))\n' % (str(mod.__name__), data)
+
+    modules = [
+        'six', 'docopt', 'ConfigObject', sys.modules[__name__]
+    ] + arguments.get('<MODULE>', [])
+    modules = ''.join([encode_module(m) for m in modules])
+
+    for name in filenames:
+        script = os.path.join(dest, name.replace('_', '-'))
+        with open(script, 'w') as fd:
+            fd.write(SCRIPT_HEADER + modules + LOAD_MODULES)
+            fd.write(inspect.getsource(mod).replace('__main__',
+                                                    '__chutified__'))
+            fd.write("if __name__ == '__main__':\n    %s()\n" % name)
+        executable = sh.chmod('+x', script)
+        if executable:
+            print(executable.commands_line)
+        else:
+            print('failed to generate %s' % script)
+    return 0
+
+
+SCRIPT_HEADER = '''
+#!/usr/bin/env python
+import base64, json, types, sys
+PY3 = sys.version_info[0] == 3
+mods = []
+'''.lstrip()
+
+LOAD_MODULES = '''
+for name, code in mods:
+    if PY3:
+        if isinstance(code, str):
+            code = code.encode('utf-8')
+    else:
+        name = bytes(name)
+    code = base64.decodestring(code)
+    mod = types.ModuleType(name)
+    globs = dict()
+    if PY3:
+        if isinstance(code, bytes):
+            code = code.decode('utf-8')
+        exec(code, globs)
+    else:
+        exec('exec code in globs')
+    mod.__dict__.update(globs)
+    if name == 'chut':
+        mod.wraps_module(mod)
+    else:
+        sys.modules[name] = mod
+
+import six
+'''.lstrip()
+
 if __name__ != '__main__':
     mod = sys.modules[__name__]
     wraps_module(mod)
